;; stboot.S
;; by trya - tryagainprod@gmail.com

;; CEFULL is loaded at 0x100000
;; in 32-bit protected mode, hence...
[ORG 0x100000]
[BITS 32]

cefull:
header:
;; Reproduce CE header (0xc0 length)
  jmp short header+0x40 ; eb3e
  db 0x45,0x54
  dd cefull_end - cefull ; executable size (checked by DeviceVM rom loader)
  times 0x38 db 0
  jmp short header+0xc0 ; eb7e
  db 0x43,0x45
  times 0x7c db 0
header_end:

stage1:
;; Our code starts here
  cli

;; Load our GDT
  lgdt [stage1_gdtr]

;; Jump to our 32-bit code segment
  jmp far 0x8:pm_code

pm_code:
;; Tell the segment registers to refer to the new data segment descriptor  
  mov eax, 0x10
  mov ds, eax
  mov ss, eax
  mov es, eax
  mov fs, eax
  mov gs, eax

;; Copy the included boot sector to 0x7c00
  cld
  mov esi, mbr
  mov edi, 0x7c00
  mov ecx, mbr_end - mbr
  rep movsb

;; Copy the 2nd stage code and data to 0x1000
  mov esi, stage2
  mov edi, 0x1000
  mov ecx, stage2_end - stage2
  rep movsb

;; Load the GDT from real mode address space
  lgdt [stage2_gdtr - stage2 + 0x1000]

;; Jump to our new code in 16-bit protected mode (16-bit GDT code selector)
  jmp far 0x18:0x1000
stage1_end:

[BITS 16]

stage2:
;; Tell the segment registers that real mode compatible values are in a 16-bit
;; data segment descriptor
  mov ax, 0x20
  mov ds, ax
  mov ss, ax
  mov es, ax
  mov fs, ax
  mov gs, ax

;; Disable protected mode flag in CR0
  mov eax, cr0
  and al, 11111110b
  mov cr0, eax

;; Reinitialize the segment registers with valid values (for example 0x0)
  mov ax, 0
  mov ds, ax
  mov ss, ax
  mov es, ax
  mov fs, ax
  mov gs, ax

;; Disable A20, assuming it was enabled by the rom loader
  in  al, 0x92
  and al, 11111101b
  out 0x92, al

;; Load real mode IDT
  lidt [idtr - stage2 + 0x1000]

;; Jump to real mode code address
  jmp far 0x0:(rm_code - stage2 + 0x1000)

rm_code:
;; Back into serious business, enable interrupts
  sti

;; Do anything you want in real mode here
  jmp far 0x0:0x7c00

;; GDT: table of 64-bit entries,
;; first element of the table is NULL.
;; Usual descriptor structure (rather complicated):
;; |___db___|___db___|___db___|___db___|___db___|___db___|____db_____|___db___|
;; |0______________15|16______________________39|40____47|48_51|52_55|56____63|
;; |      Limit      |                          |        |Limit|     |  Base  |
;; |      first      |    Base first 24 bits    | Flags  |last |Flags|  last  |
;; |_____16 bits_____|__________________________|________|4bits|_____|_8_bits_|

gdt:
  db 0,0,0,0,0,0,0,0
gdt_pm_cs:
  db 0xFF,0xFF,0x0,0x0,0x0,10011010b,11001111b,0x0
gdt_pm_ds:
  db 0xFF,0xFF,0x0,0x0,0x0,10010010b,11001111b,0x0
gdt_rm_cs:
  db 0xFF,0xFF,0x0,0x0,0x0,10011010b,00000000b,0x0
gdt_rm_ds:
  db 0xFF,0xFF,0x0,0x0,0x0,10010010b,00000000b,0x0
gdt_end:

;; GDT register: 48-bit (16+32) register
;; |_____dw_____|__________dd__________|
;; |0_________15|16__________________47|
;; |            |                      |
;; |  GDT size  |     GDT pointer      |
;; |____________|______________________|

stage1_gdtr:
  dw gdt_end - gdt - 1
  dd gdt

stage2_gdtr:
  dw gdt_end - gdt - 1
  dd gdt - stage2 + 0x1000

;; IDT register: same structure as GDTR

idtr:
  dw 0x3ff ; 256 entries, 4b each = 1K
  dd 0     ; Real Mode IVT @ 0x0000
stage2_end:

;; Included boot sector, to be copied to 0x7c00
mbr:
  INCBIN "bootsect.bin"
mbr_end:
cefull_end:
